from collections import defaultdict, Sequence, Sized, Iterable, Callable
import inspect
import wrapt
from cytoolz import curry
from numpy import ndarray

from .exceptions import UndefinedOperatorError, DifferentLengthError
from .exceptions import ExpectedTypeError, ShapeMismatchError
from .exceptions import OutsideRangeError


def equal_length(*args):
    '''Check that argument have the same length.
    '''
    for i, arg in enumerate(args):
        if not isinstance(arg, Sized):
            raise ExpectedTypeError(arg, ['Sized'])
        if i >= 1 and len(arg) != len(args[0]):
            raise DifferentLengthError(args, arg)


@curry
def has_shape(shape, arg_id, args, kwargs):
    '''Check that a particular argument is an array with a given shape. The
    shape may contain string attributes, which will be fetched from arg0 to
    the function (usually self).
    '''
    self = args[0]
    arg = args[arg_id]
    if not hasattr(arg, 'shape'):
        raise ExpectedTypeError(arg, ['array'])
    shape_values = []
    for dim in shape:
        if not isinstance(dim, int):
            dim = getattr(self, dim, None)
        shape_values.append(dim)
    if len(shape) != len(arg.shape):
        raise ShapeMismatchError(arg.shape, tuple(shape_values), shape)
    for i, dim in enumerate(shape_values):
        # Allow underspecified dimensions
        if dim != None and arg.shape[i] != dim:
            raise ShapeMismatchError(arg.shape, shape_values, shape)


def is_shape(arg_id, args, func_kwargs, **kwargs):
    arg = args[arg_id]
    if not isinstance(arg, Iterable):
        raise ExpectedTypeError(arg, ['iterable'])
    for value in arg:
        if not isinstance(value, int) or value < 0:
            raise ExpectedTypeError(arg, ['valid shape (positive ints)'])


def is_sequence(arg_id, args, kwargs):
    arg = args[arg_id]
    if not isinstance(arg, Iterable):
        raise ExpectedTypeError(arg, ['iterable'])


def is_float(arg_id, args, func_kwargs, **kwargs):
    arg = args[arg_id]
    if not isinstance(arg, float):
        raise ExpectedTypeError(arg, ['float'])
    if 'min' in kwargs and arg < kwargs['min']:
        raise OutsideRangeError(arg, kwargs['min'], '>=')
    if 'max' in kwargs and arg > kwargs['max']:
        raise OutsideRangeError(arg, kwargs['max'], '<=')


def is_int(arg_id, args, func_kwargs, **kwargs):
    arg = args[arg_id]
    if not isinstance(arg, int):
        raise ExpectedTypeError(arg, ['int'])
    if 'min' in kwargs and arg < kwargs['min']:
        raise OutsideRangeError(arg, kwargs['min'], '>=')
    if 'max' in kwargs and arg > kwargs['max']:
        raise OutsideRangeError(arg, kwargs['max'], '<=')


def is_array(arg_id, args, func_kwargs, **kwargs):
    arg = args[arg_id]
    if not isinstance(arg, ndarray):
        raise ExpectedTypeError(arg, ['ndarray'])


def is_int_array(arg_id, args, func_kwargs, **kwargs):
    arg = args[arg_id]
    if not isinstance(arg, ndarray) or 'i' not in arg.dtype.kind:
        raise ExpectedTypeError(arg, ['ndarray[int]'])


def operator_is_defined(op):
    @wrapt.decorator
    def checker(wrapped, instance, args, kwargs):
        if instance is None:
            instance = args[0]
        if instance is None:
            raise ExpectedTypeError(instance, ['Model'])
        if op not in instance._operators:
            raise UndefinedOperatorError(op, instance, args[0], instance._operators)
        else:
            return wrapped(*args, **kwargs)
    return checker


def arg(arg_id, *constraints):
    @wrapt.decorator
    def checked_function(wrapped, instance, args, kwargs):
        # for partial functions or other C-compiled functions
        if not hasattr(wrapped, 'checks'): # pragma: no cover
            return wrapped(*args, **kwargs)
        if instance is not None:
            fix_args = [instance] + list(args)
        else:
            fix_args = list(args)
        for arg_id, checks in wrapped.checks.items():
            for check in checks:
                if not isinstance(check, Callable):
                    raise ExpectedTypeError(check, ['Callable'])
                check(arg_id, fix_args, kwargs)
        return wrapped(*args, **kwargs)

    def arg_check_adder(func):
        if hasattr(func, 'checks'):
            func.checks.setdefault(arg_id, []).extend(constraints)
            return func
        else:
            wrapped = checked_function(func)
            wrapped.checks = {arg_id: list(constraints)}
            return wrapped
    return arg_check_adder


def args(*constraints):
    @wrapt.decorator
    def arg_check_adder(wrapped, instance, args, kwargs):
        for check in constraints:
            if not isinstance(check, Callable):
                raise ExpectedTypeError(check, ['Callable'])
            check(*args)
        return wrapped(*args, **kwargs)
    return arg_check_adder



from __future__ import unicode_literals, print_function

from .optimizers import Eve, Adam, SGD, linear_decay
from .util import minibatch

import numpy.random
import tqdm


class Trainer(object):
    def __init__(self, model, train_data, L2=0.0):
        self.ops = model.ops
        self.model = model
        self.optimizer = Eve(Adam(model.ops, 0.001))
        self.batch_size = 128
        self.nb_epoch = 1
        self.i = 0
        self.L2 = 0.0
        self.dropout = 0.0
        self.dropout_decay = 0.0
        self._loss = 0.
        self.each_epoch = []

    def __enter__(self):
        return self, self.optimizer

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.model.use_params(self.optimizer.averages)

    def iterate(self, train_X, train_y):
        orig_dropout = self.dropout
        for i in range(self.nb_epoch):
            indices = self.ops.xp.asarray(range(len(train_X)))
            numpy.random.shuffle(indices)
            j = 0
            while j < len(indices):
                slice_ = indices[j : j + self.batch_size]
                X = _take_slice(train_X, slice_)
                y = _take_slice(train_y, slice_)
                yield X, y
                self.dropout = linear_decay(orig_dropout, self.dropout_decay,
                                            self.optimizer.nr_iter)
                j += self.batch_size
            for func in self.each_epoch:
                func()


def _take_slice(data, slice_):
    x = [data[i] for i in slice_]
    return x




from libc.stdint cimport int16_t, int32_t, uint64_t
from preshed.maps cimport MapStruct
from libcpp.vector cimport vector
from libc.stdlib cimport malloc, calloc, free, realloc
from libc.string cimport memcpy, memset
from murmurhash.mrmr cimport real_hash64 as hash64

from .typedefs cimport len_t, idx_t, atom_t, weight_t
from .linalg cimport VecVec


include "compile_time_constants.pxi"

ctypedef vector[weight_t] vector_weight_t


ctypedef void (*do_update_t)(
    weight_t* weights,
    weight_t* gradient,
        len_t nr,
        const ConstantsC* hp,
) nogil


ctypedef void (*do_feed_fwd_t)(
    weight_t** fwd,
        const weight_t* W,
        const len_t* shape,
        int nr_layer,
        int nr_batch,
        const ConstantsC* hp
) nogil
 

ctypedef void (*do_feed_bwd_t)(
    weight_t* G,
    weight_t** bwd,
        const weight_t* W,
        const weight_t* const* fwd,
        const len_t* shape,
        int nr_layer,
        int nr_batch,
        const ConstantsC* hp
) nogil


# Alias this, so that it matches our naming scheme
ctypedef MapStruct MapC


cdef struct ConstantsC:
    weight_t a
    weight_t b
    weight_t c
    weight_t d
    weight_t e
    weight_t g
    weight_t h
    weight_t i
    weight_t j
    weight_t k
    weight_t l
    weight_t m
    weight_t n
    weight_t o
    weight_t p
    weight_t q
    weight_t r
    weight_t s
    weight_t t
    weight_t u
    weight_t w
    weight_t x
    weight_t y
    weight_t z


cdef struct EmbedC:
    MapC** weights
    MapC** gradients
    weight_t** defaults
    weight_t** d_defaults
    idx_t* offsets
    len_t* lengths
    len_t nr
    int nr_support


cdef struct NeuralNetC:
    do_feed_fwd_t feed_fwd
    do_feed_bwd_t feed_bwd
    do_update_t update

    len_t* widths
    weight_t* weights
    weight_t* gradient

    EmbedC* embed

    len_t nr_layer
    len_t nr_weight
    len_t nr_node

    ConstantsC hp


cdef extern from "stdlib.h":
    int posix_memalign(void **memptr, size_t alignment, size_t size) nogil
    void* valloc (size_t size) nogil


cdef struct ExampleC:
    int* is_valid
    weight_t* costs
    uint64_t* atoms
    FeatureC* features
    weight_t* scores

    int nr_class
    int nr_atom
    int nr_feat


cdef cppclass MinibatchC:
    weight_t** _fwd
    weight_t** _bwd
    
    FeatureC** _feats
    len_t* _nr_feat
    
    weight_t* _costs
    int* _is_valid
    uint64_t* signatures

    len_t* widths
    int i
    int nr_layer
    int batch_size

    __init__(len_t* widths, int nr_layer, int batch_size) nogil:
        this.i = 0
        this.nr_layer = nr_layer
        this.batch_size = batch_size
        this.widths = <len_t*>calloc(nr_layer, sizeof(len_t))
        this._fwd = <weight_t**>calloc(nr_layer, sizeof(weight_t*))
        this._bwd = <weight_t**>calloc(nr_layer, sizeof(weight_t*))
        for i in range(nr_layer):
            this.widths[i] = widths[i]
            this._fwd[i] = <weight_t*>calloc(this.widths[i] * batch_size, sizeof(weight_t))
            this._bwd[i] = <weight_t*>calloc(this.widths[i] * batch_size, sizeof(weight_t))
        this._feats = <FeatureC**>calloc(batch_size, sizeof(void*))
        this._nr_feat = <len_t*>calloc(batch_size, sizeof(len_t))
        this._is_valid = <int*>calloc(batch_size * widths[nr_layer-1], sizeof(int))
        this._costs = <weight_t*>calloc(batch_size * widths[nr_layer-1], sizeof(weight_t))
        this.signatures = <uint64_t*>calloc(batch_size, sizeof(uint64_t))

    __dealloc__() nogil:
        free(this.widths)
        for i in range(this.nr_layer):
            free(this._fwd[i])
            free(this._bwd[i])
        for i in range(this.i):
            free(this._feats[i])
        free(this._fwd)
        free(this._bwd)
        free(this._feats)
        free(this._nr_feat)
        free(this._is_valid)
        free(this._costs)
        free(this.signatures)

    void reset() nogil:
        for i in range(this.nr_layer):
            memset(this._fwd[i],
                0, sizeof(this._fwd[i][0]) * this.batch_size * this.widths[i])
            memset(this._bwd[i],
                0, sizeof(this._bwd[i][0]) * this.batch_size * this.widths[i])
        memset(this._nr_feat, 0, sizeof(this._nr_feat[0]) * this.batch_size)
        memset(this.signatures, 0, sizeof(this.signatures[0]) * this.batch_size)
        memset(this._costs,
            0, sizeof(this._costs[0]) * this.nr_out() * this.batch_size)
        memset(this._is_valid,
            0, sizeof(this._is_valid[0]) * this.nr_out() * this.batch_size)
        for i in range(this.i):
            free(this._feats[i])
            this._feats[i] = NULL
        this.i = 0

    int nr_in() nogil:
        return this.widths[0]

    int nr_out() nogil:
        return this.widths[this.nr_layer - 1]

    int push_back(const FeatureC* feats, int nr_feat,
            const weight_t* costs, const int* is_valid, uint64_t key) nogil:
        # Hash the features, to see if the batch has a matching input.
        # If it does, just update the gradient for it.
        if key != 0:
            for i in range(this.i):
                if this.signatures[i] == key:
                    VecVec.add_i(this.costs(i),
                        costs, 1.0, this.nr_out())
                    return 0
        if this.i >= this.batch_size:
            this.reset()
            this.i = 0 # This is done in reset() --- but make it obvious
 
        this.signatures[this.i] = key
        this._nr_feat[this.i] = nr_feat
        this._feats[this.i] = <FeatureC*>calloc(nr_feat, sizeof(FeatureC))
        memcpy(this._feats[this.i],
            feats, nr_feat * sizeof(this._feats[this.i][0]))

        memcpy(this.costs(this.i),
            costs, this.nr_out() * sizeof(costs[0]))
        if is_valid is not NULL:
            memcpy(this.is_valid(this.i),
                is_valid, this.nr_out() * sizeof(is_valid[0]))
        else:
            for i in range(this.nr_out()):
                this.is_valid(this.i)[i] = 1
        this.i += 1
        return this.i >= this.batch_size

    FeatureC* features(int i) nogil:
        return this._feats[i]

    int nr_feat(int i) nogil:
        return this._nr_feat[i]

    weight_t* fwd(int i, int j) nogil:
        return this._fwd[i] + (j * this.widths[i])
 
    weight_t* bwd(int i, int j) nogil:
        return this._bwd[i] + (j * this.widths[i])
  
    weight_t* scores(int i) nogil:
        return this.fwd(this.nr_layer-1, i)

    weight_t* losses(int i) nogil:
        return this.bwd(this.nr_layer-1, i)
 
    weight_t* costs(int i) nogil:
        return this._costs + (i * this.nr_out())

    int* is_valid(int i) nogil:
        return this._is_valid + (i * this.nr_out())

    int guess(int i) nogil:
        return VecVec.arg_max_if_true(this.scores(i), this.is_valid(i), this.nr_out())
    
    int best(int i) nogil:
        return VecVec.arg_max_if_zero(this.scores(i), this.costs(i), this.nr_out())



cdef packed struct SparseArrayC:
    int32_t key
    weight_t val


cdef struct FeatureC:
    int i
    uint64_t key
    weight_t value


cdef struct SparseAverageC:
    SparseArrayC* curr
    SparseArrayC* mom1
    SparseArrayC* mom2
    SparseArrayC* avgs
    SparseArrayC* times
    weight_t penalty


cdef struct TemplateC:
    int[MAX_TEMPLATE_LEN] indices
    int length
    atom_t[MAX_TEMPLATE_LEN] atoms



import numpy as np
from .model import Model
from ... import describe
from ... import check
from ...check import is_int_array, is_int
from ... describe import Dimension, Weights, Synapses, Gradient
from .._lsuv import svd_orthonormal, do_lsuv


def _set_dimensions_if_needed(model, X, y=None):
    if model.nV == None:
        max_id = int(X.max()) + 1
        if max_id >= 10000000: # pragma: no cover
            raise ValueError("TODO error --- really want us to make 1m vectors?")
        model.nV = max_id


def _uniform_init(lo, hi):
    def wrapped(W, ops):
        W[:] = ops.xp.random.uniform(lo, hi, W.shape)
    return wrapped


def LSUVinit(model, X, y=None):
    if model.vectors is not None and model.W is not None:
        do_lsuv(model.ops, model.vectors, model, X)
        do_lsuv(model.ops, model.W, model, X)
    return X


@describe.on_data(_set_dimensions_if_needed, LSUVinit)
@describe.attributes(
    nM=Dimension("Vector dimensions"),
    nV=Dimension("Number of vectors"),
    nO=Dimension("Size of output"),
    W=Synapses(
        "A projection matrix, to change vector dimensionality",
        lambda obj: (obj.nO, obj.nM),
        lambda W, ops: ops.xavier_uniform_init(W)),
    vectors=Weights("Embedding table",
        lambda obj: (obj.nV, obj.nM),
        _uniform_init(-0.1, 0.1)
    ),
    d_W=Gradient("W"),
    d_vectors=Gradient("vectors")
)
class Embed(Model):
    name = 'embed'

    #@property
    #def input_shape(self):
    #    return (self.nB,)

    #@property
    #def output_shape(self):
    #    return (self.nB, self.nO)

    @check.arg(1, is_int)
    def __init__(self, nO, nM=None, nV=None, **kwargs):
        Model.__init__(self, **kwargs)
        self.nO = nO
        self.nM = nM
        self.nV = nV
        self._id_map = {0: 0}

    @check.arg(1, is_int_array)
    def predict(self, ids):
        if len(ids) < 1000:
            vectors = self._embed(ids)
            return self.ops.batch_dot(vectors, self.W)
        id_map = {}
        for i, id_ in enumerate(ids):
            if id_ not in id_map:
                id_map[id_] = [i]
            else:
                id_map[id_].append(i)
        mapped = sorted(id_map.items())
        vectors = self._embed([id_ for id_, _ in mapped])
        result = self.ops.batch_dot(vectors, self.W)
        output = self.ops.allocate((len(ids), self.nO))
        for i, (_, occurs) in enumerate(mapped):
            for j in occurs:
                output[j] = result[i]
        return output

    def begin_update(self, ids, drop=0.):
        def finish_update(gradients, sgd=None):
            self.d_W += self.ops.batch_outer(gradients, self._embed(ids))
            gradients = self.ops.batch_dot(gradients, self.W.T)
            d_vectors = self.d_vectors
            n_vector = d_vectors.shape[0]
            for id_, delta_in in zip(ids, gradients):
                id_ = self._id_map[id_]
                if id_ < n_vector:
                    d_vectors[id_] += delta_in
            if sgd is not None:
                sgd(self._mem.weights, self._mem.gradient, key=id(self._mem))
            return None
        return self.predict(ids), finish_update

    def _embed(self, ids):
        n_vector = len(self._id_map)
        for id_ in ids:
            if id_ not in self._id_map:
                self._id_map[id_] = n_vector
                n_vector += 1
        output = self.ops.allocate((len(ids), self.nM))
        weights = self.vectors 
        n_vector = weights.shape[0]
        for i, id_ in enumerate(ids):
            id_ = self._id_map[id_]
            if id_ < n_vector:
                output[i] = weights[id_]
        return output



from collections import defaultdict

from .model import Model
from .._lsuv import do_lsuv
from ... import describe
from ...describe import Dimension, Synapses, Biases, Gradient
from ... import check
from ...check import is_sequence


def _set_dimensions_if_given(model, *args, **kwargs):
    if len(args) >= 1:
        model.nO = args[0]
    if 'pieces' in kwargs:
        model.nP = kwargs['pieces']
    if 'window' in kwargs:
        model.nF = kwargs['window'] * 2 + 1
    model.nI = model.embed.nO


def _set_dimensions_if_needed(model, X, y=None):
    if model.nO == None and y is not None:
        model.nO = y.max() + 1
 

def LSUVinit(model, positions, y=None):
    ids = []
    for id_, occurs in positions.items():
        ids.extend(id_ for _ in occurs)
    ids = model.ops.asarray(ids, dtype='i')
    for hook in model.embed.on_data_hooks:
        hook(model.embed, ids, y)
    return do_lsuv(model.ops, model.W, model, positions)


@describe.input(("nB", "nI"))
@describe.output(("nB", "nO"))
@describe.on_data(_set_dimensions_if_needed, LSUVinit)
@describe.on_init(_set_dimensions_if_given)
@describe.attributes(
    nP=Dimension("Number of pieces"),
    nF=Dimension("Number of features"),
    nO=Dimension("Size of output"),
    nI=Dimension("Size of input"),
    W=Synapses("Weights matrix", lambda obj: (obj.nO, obj.nP, obj.nF, obj.nI),
        lambda W, ops: ops.xavier_uniform_init(W)),
    b=Biases("Bias vector", lambda obj: (obj.nO, obj.nP)),
    d_W=Gradient("W"),
    d_b=Gradient("b")
)
class MaxoutWindowEncode(Model):
    name = 'window-encode'
    @property
    def nW(self):
        return int((self.nF-1)/2)

    def __init__(self, embed, *args, **kwargs):
        self.embed = embed
        Model.__init__(self, *args, **kwargs)

    def predict(self, positions):
        uniq_ids = self.ops.asarray(sorted(positions.keys()), dtype='i')
        uniq_vectors = self.embed.predict(uniq_ids)
        out, _ = self._forward(uniq_ids, positions, uniq_vectors)
        return out

    def begin_update(self, positions, drop=0.):
        uniq_ids = self.ops.asarray(sorted(positions.keys()))
        uniq_vectors, fine_tune = self.embed.begin_update(uniq_ids, drop=drop)
        assert uniq_ids.shape[0] == uniq_vectors.shape[0]
        
        best__bo, which__bo = self._forward(uniq_ids, positions, uniq_vectors)
        best__bo, bp_dropout = self.ops.dropout(best__bo, drop, inplace=True)

        def finish_update(gradient__bo, sgd=None):
            gradient__bop = self.ops.backprop_maxout(gradient__bo, which__bo, self.nP)
            self.d_b += gradient__bop.sum(axis=0)
            inputs__bfi = _get_full_inputs(
                self.ops, uniq_ids, positions, uniq_vectors, self.nW)
            if fine_tune is not None:
                gradient__bfi = self.ops.xp.tensordot(gradient__bop, self.W,
                    axes=[[1,2], [0,1]])
                gradient__ui = _get_vector_gradients(self.ops, uniq_ids, positions,
                                                     gradient__bfi)
                grad_out = fine_tune(gradient__ui)
            else:
                grad_out = None
            # Bop,Bfi->opfi
            self.d_W += self.ops.xp.tensordot(gradient__bop, inputs__bfi, axes=[[0], [0]])
            if sgd is not None: # pragma: no cover
                sgd(self._mem.weights, self._mem.gradient, key=id(self._mem))
            return grad_out
        return best__bo, bp_dropout(finish_update)

    def _forward(self, uniq_ids, positions, vectors):
        assert self.nP != 0
        hidden = _compute_hidden_layer(self.ops, self.W, vectors)
        cands = _get_output(self.ops, uniq_ids, positions, hidden)
        cands += self.b
        return self.ops.maxout(cands)


def _get_output(ops, uniq_ids, positions, H__ufop):
    nU, nF, nO, nP = H__ufop.shape
    nW = int((nF-1)/2)
    total_length = sum(len(val) for val in positions.values())
    # Shift the input, so that we don't have to special-case the starts and
    # ends. We'll shift back afterwards.
    out__bop = ops.allocate((total_length+(nW*2), nO, nP))
    # Let's say we have the word 'of'
    # Its vector is at vec_idx=1. It occurred at [1, 10, 12, 17] in our data.
    # Our data has 17 tokens.
    for vec_idx, id_ in enumerate(uniq_ids):
        tok_idxs = positions[id_] # [1, 10, 12, 17]
        # v__fop holds the feature weights for 'of' in 5 rows:
        # 0: _ _ _ _ of
        # 1: _ _ _ of _
        # 2: _ _ of _ _
        # 3: _ of _ _ _
        # 4: of _ _ _ _
        # The weights at row 0 are the ones that apply for a word two before 'of',
        # i.e. when 'of' is in the RR position. The weights at row 4 are the ones
        # that apply when 'of' is two *after* the focus word. We can therefore
        # add these weights to a slice of the output.
        v__fop = H__ufop[vec_idx]
        ops.increment_slices(out__bop, v__fop, tok_idxs)
    # Shift the output, to correct for the 'padding' shift above.
    out__bop = out__bop[nW : -nW]
    return out__bop


def _compute_hidden_layer(ops, W__opfi, vectors__ui):
    H__uopf = ops.xp.tensordot(vectors__ui, W__opfi, axes=[[1], [3]])
    H__ufop = H__uopf.transpose((0, 3, 1, 2))
    return H__ufop


def _get_vector_gradients(ops, uniq_ids, positions, tune__bfi):
    nW = int((tune__bfi.shape[1]-1) / 2)
    gradients__ui = ops.allocate((len(uniq_ids), tune__bfi.shape[-1]))
    for u, id_ in enumerate(uniq_ids):
        d_vector = gradients__ui[u]
        for i in positions[id_]:
            # Let's say the input was
            # "the heart of the matter"
            # What's the gradient for the vector of "of"?
            # It was used 5 times --- as a feature for each word
            for f in range(nW):
                if i-(nW-f) >= 0:
                    d_vector += tune__bfi[i-(nW-f), f]
            for f in range(nW+1):
                if (i+f) < tune__bfi.shape[0]:
                    d_vector += tune__bfi[i+f, nW+f]
    return gradients__ui


def _get_full_inputs(ops, uniq_ids, positions, vectors__ui, nW):
    total_length = sum(len(occurs) for occurs in positions.values())
    vectors__bi = ops.allocate((total_length, vectors__ui.shape[1]))
    # Get the (non-unique) vectors, as a contiguous array.
    for vec_idx, id_ in enumerate(uniq_ids):
        tok_idxs = positions[id_]
        vector = vectors__ui[vec_idx]
        for tok_idx in tok_idxs:
            vectors__bi[tok_idx] = vector
    # Let's say the input was
    # "the heart of the matter"
    # If "of" is RR, we read of[0]
    # If "of" is R, we read of[1]
    # If "of" is W, we read of[2]
    # If "of" is L, we read of[3]
    # If "of" is LL, we read of[4]
    X__bfi = ops.allocate((total_length, nW*2+1, vectors__ui.shape[1]))
    for f in range(1, nW+1):
        X__bfi[:-f, nW-f] = vectors__bi[f:] # Words at the start aren't R features
    X__bfi[:, nW] = vectors__bi
    for f in range(1, nW+1):
        X__bfi[f:, nW+f] = vectors__bi[:-f] # Words at the end aren't L features
    return X__bfi



from libc.stdint cimport uint64_t
from cymem.cymem cimport Pool

from murmurhash.mrmr cimport hash64

from ..extra.eg cimport Example

include "../compile_time_constants.pxi"


cdef class ConjunctionExtracter:
    """Extract composite features from a sequence of atomic values, according to
    the schema specified by a list of templates.
    """
    def __init__(self, templates, linear_mode=True):
        self.mem = Pool()
        nr_atom = 0
        for templ in templates:
            nr_atom = max(nr_atom, max(templ))
        self.linear_mode = linear_mode
        self.nr_atom = nr_atom
        # Value that indicates the value has been "masked", e.g. it was pruned
        # as a rare word. If a feature contains any masked values, it is dropped.
        templates = tuple(sorted(set([tuple(sorted(f)) for f in templates])))
        self._py_templates = templates
        self.nr_embed = 1
        self.nr_templ = len(templates) + 1
        self.templates = <TemplateC*>self.mem.alloc(len(templates), sizeof(TemplateC))
        # Sort each feature, and sort and unique the set of them
        cdef int i, j, idx
        for i, indices in enumerate(templates):
            assert len(indices) < MAX_TEMPLATE_LEN
            for j, idx in enumerate(sorted(indices)):
                self.templates[i].indices[j] = idx
            self.templates[i].length = len(indices)

    def __call__(self, Example eg):
        eg.c.nr_feat = self.set_features(eg.c.features, eg.c.atoms)

    cdef int set_features(self, FeatureC* feats, const atom_t* atoms) nogil:
        cdef int n_feats = 0
        if self.linear_mode:
            feats[0].key = 1
            feats[0].value = 1
            n_feats += 1
        cdef bint seen_non_zero
        cdef int templ_id
        cdef int i
        cdef atom_t[MAX_TEMPLATE_LEN] extracted
        for templ_id in range(self.nr_templ-1):
            templ = self.templates[templ_id]
            if not self.linear_mode and templ.length == 1:
                feats[n_feats].key = atoms[templ.indices[0]]
                feats[n_feats].value = 1
                feats[n_feats].i = templ_id
                n_feats += 1
                continue
            seen_non_zero = False
            for i in range(templ.length):
                extracted[i] = atoms[templ.indices[i]]
                seen_non_zero = seen_non_zero or extracted[i]
            if seen_non_zero:
                feats[n_feats].key = hash64(extracted, templ.length * sizeof(extracted[0]),
                                            templ_id if self.linear_mode else 0)
                feats[n_feats].value = 1
                feats[n_feats].i = templ_id
                n_feats += 1
        return n_feats

    def __reduce__(self):
        return (self.__class__, (self.nr_atom, self._py_templates))



from numpy import prod


from .. import check
from ..check import is_shape


class Memory(object):
    def __init__(self, ops, size=128):
        if size < 0:
            raise ValueError("TODO error re negative size %d" % size)
        self.ops = ops
        self._mem = self.ops.allocate((2, size))
        self._offsets = {}
        self._i = 0

    @property
    def weights(self):
        return self._mem[0, :self._i]

    @property
    def gradient(self):
        return self._mem[1, :self._i]

    def __contains__(self, name):
        return name in self._offsets

    def __getitem__(self, name):
        offset, col, shape = self._offsets[name]
        return self._mem[col, offset : offset + prod(shape)].reshape(shape)

    def get(self, name, default=None):
        return self[name] if name in self._offsets else default

    def set(self, value):
        self._mem[0, :self._i] = value

    #@check.arg(2, is_shape)
    def add(self, name, shape):
        assert name not in self._offsets, "TODO error"
        self._offsets[name] = (self._i, 0, shape)
        blob = self._get_blob(prod(shape))
        return blob[0].reshape(shape)
    
    def add_gradient(self, grad_name, param_name):
        assert grad_name not in self._offsets, "TODO error"
        offset, _, shape = self._offsets[param_name]
        self._offsets[grad_name] = (offset, 1, shape)
        return self._mem[1, offset : offset + prod(shape)].reshape(shape)
    
    def _get_blob(self, nr_req):
        nr_avail = self._mem.shape[1] - (self._i+1)
        if nr_avail < nr_req:
            self._realloc(max(self._mem.shape[1], nr_req) * 2)
        blob = self._mem[:, self._i : self._i + nr_req]
        self._i += nr_req
        return blob

    def _realloc(self, new_size):
        new_mem = self.ops.allocate((self._mem.shape[0], new_size))
        new_mem[:, :self._i+1] = self._mem[:, :self._i+1]
        self._mem = new_mem

#
#    def merge_params(self, others):
#        others = list(others)
#        if not others:
#            return None
#        if not all(other.allow_resize for other in others):
#            raise ValueError("TODO Error")
#        sizes = [other._i+1 for other in others]
#        nr_req = self._i + sum(sizes)
#        if self._mem.shape[1] < nr_req:
#            self._realloc(nr_req)
#        self.allow_resize = False
#        for other in others:
#            other.replace_mem(self._get_blob(other._i))
#    
#    def replace_mem(self, mem):
#        if not self.allow_resize:
#            raise ValueError("TODO Error")
#        self.allow_resize = False
#        mem[:] = self._mem[:, :self._i]
#        self._mem = mem
#
#



# coding: utf-8
from __future__ import unicode_literals
from collections import Sized

import traceback
from termcolor import colored as color


class UndefinedOperatorError(TypeError):
    def __init__(self, op, arg1, arg2, operators):
        self.tb = traceback.extract_stack()
        TypeError.__init__(self, get_error(
            "Undefined operator: {op}".format(op=op),
            "Called by ({arg1}, {arg2})".format(arg1=arg1, arg2=arg2),
            "Available: {ops}".format(ops= ', '.join(operators.keys())),
            tb=self.tb,
            highlight=op
        ))


class OutsideRangeError(ValueError):
    def __init__(self, arg, val, operator):
        self.tb = traceback.extract_stack()
        ValueError.__init__(self, get_error(
            "Outside range: {v} needs to be {o} {v2}".format(v=arg, o=operator, v2=val),
            tb=self.tb
        ))


class DifferentLengthError(ValueError):
    def __init__(self, args, arg):
        self.tb = traceback.extract_stack()
        vals = ['{v} [{l}]'.format(v=arg, l=len(arg) if isinstance(arg, Sized) else 'no length') for arg in args]
        ValueError.__init__(self, get_error(
            "Values need to be equal length: {v}".format(v=', '.join(vals)),
            tb=self.tb
        ))


class ShapeMismatchError(ValueError):
    def __init__(self, shape, dim, shape_names):
        self.tb = traceback.extract_stack()
        ValueError.__init__(self, get_error(
            "Shape mismatch: input {s} not compatible with {d}.".format(s=shape, d=dim),
            tb=self.tb
        ))


class ExpectedTypeError(TypeError):
    max_to_print_of_value = 200
    def __init__(self, bad_type, expected):
        self.tb = traceback.extract_stack()
        bad_type = repr(bad_type)
        if len(bad_type) >= self.max_to_print_of_value:
            half = int(self.max_to_print_of_value/2)
            bad_type = bad_type[:half] + ' ... ' + bad_type[-half:]
        TypeError.__init__(self, get_error(
            "Expected type {e}, but got: {v} ({t})".format(e='/'.join(expected), v=bad_type, t=type(bad_type)),
            tb=self.tb,
            highlight=bad_type
        ))


def get_error(title, *args, **kwargs):
    template = '\n\n\t{title}{info}{tb}\n'
    info = '\n'.join(['\t' + l for l in args]) if args else ''
    highlight = kwargs['highlight'] if 'highlight' in kwargs else False
    tb = _get_traceback(kwargs['tb'], highlight) if 'tb' in kwargs else ''
    return template.format(title=color(title, 'red', attrs=['bold']),
                           info=info, tb=tb).encode('utf8')


def _get_traceback(tb, highlight):
    template = '\n\n\t{title}\n\t{tb}'
    # Prune "check.py" from tb (hacky)
    tb = [record for record in tb if not record[0].endswith('check.py')]
    tb_range = tb[-5:-2]
    tb_list = [_format_traceback(p, l, fn, t, i, len(tb_range), highlight) for i, (p, l, fn, t) in enumerate(tb_range)]
    return template.format(title=color('Traceback:', 'blue', attrs=['bold']),
                           tb='\n'.join(tb_list).strip())


def _format_traceback(path, line, fn, text, i, count, highlight):
    template = '\t{i} {fn} [{l}] in {p}{t}'
    indent = ('└─' if i == count-1 else '├─') + '──'*i
    filename = path.rsplit('/thinc/', 1)[1] if '/thinc/' in path else path
    text = _format_user_error(text, i, highlight) if i == count-1 else ''
    return template.format(l=str(line), i=indent, t=text,
                           fn=color(fn, attrs=['bold']),
                           p=color(filename, attrs=['underline']))


def _format_user_error(text, i, highlight):
    template = '\n\t  {sp} {t}'
    spacing = '  '*i + color(' >>>', 'red')
    if highlight:
        text = text.replace(str(highlight), color(str(highlight), 'yellow'))
    return template.format(sp=spacing, t=text)



import copy

from .neural._classes.feed_forward import FeedForward
from .neural._classes.model import Model


def layerize(begin_update=None, *args, **kwargs):
    '''Wrap a function into a layer'''
    if begin_update is not None:
        return FunctionLayer(begin_update, *args, **kwargs)
    def wrapper(begin_update):
        return FunctionLayer(begin_update, *args, **kwargs)
    return wrapper


def metalayerize(user_func):
    '''Wrap a function over a sequence of layers and an input into a layer.'''
    def returned(layers, *args, **kwargs):
        def begin_update(X, *args, **kwargs):
            return user_func(layers, X, *args, **kwargs)
        return FunctionLayer(begin_update, *args, **kwargs)
    return returned


def noop(*layers):
    '''Transform a sequences of layers into a null operation.'''
    def begin_update(X):
        return X, lambda D, *a, **k: D
    return begin_update


def chain(*layers):
    '''Compose two models `f` and `g` such that they become layers of a single
    feed-forward model that computes `g(f(x))`.
    
    Raises exception if their dimensions don't match.
    '''
    if len(layers) == 0:
        return FeedForward([])
    elif len(layers) == 1:
        return layers[0]
    else:
        return FeedForward(layers)


def clone(orig, n):
    '''Construct `n` copies of a layer, with distinct weights.
    
    i.e. `clone(f, 3)(x)` computes `f(f'(f''(x)))`.
    '''
    layers = [orig]
    for i in range(n-1):
        layers.append(copy.deepcopy(orig))
    return FeedForward(layers)


def concatenate(*layers): # pragma: no cover
    '''Compose two or more models `f`, `g`, etc, such that their outputs are
    concatenated, i.e. `concatenate(f, g)(x)` computes `hstack(f(x), g(x))`
    '''
    if not layers:
        return noop()
    ops = layers[0].ops
    def begin_update(X, *a, **k):
        forward, backward = split_backward(layers)
        values = [fwd(X, *a, **k) for fwd in forward]
       
        output = ops.xp.hstack(values)
        shapes = [val.shape for val in values]

        def finish_update(gradient, *args, **kwargs):
            layer_grads = []
            start = 0
            for bwd, shape in zip(backward, shapes):
                end = start + shape[1]
                if bwd is not None:
                    d = bwd(gradient[:, start : end], *args, **kwargs)
                    if d is not None:
                        layer_grads.append(d)
                start = end
            return ops.asarray(ops.xp.sum(layer_grads, axis=0))
        return output, finish_update
    nO = sum(layer.output_shape[1] for layer in layers) 
    layer = FunctionLayer(begin_update, nO=nO)
    return layer


def split_backward(layers): # pragma: no cover
    '''Separate a sequence of layers' `begin_update` methods into two lists of
    functions: one that computes the forward values, and the other that completes
    the backward pass. The backward sequence is only populated after the forward
    functions have been applied.
    '''
    backward = []
    forward = [sink_return(op.begin_update, backward.append)
               for op in layers]
    return forward, backward


def sink_return(func, sink, splitter=None): # pragma: no cover
    '''Transform a function `func` that returns tuples into a function that returns
    single values. Call a function `sink` on the unused values.
    '''
    def wrap(*args, **kwargs):
        output = func(*args, **kwargs)
        if splitter is None:
            to_keep, to_sink = output
        else:
            to_keep, to_sink = splitter(*output)
        sink(to_sink)
        return to_keep
    return wrap


class FunctionLayer(Model):
    '''Wrap functions into weightless Model instances, for use as network
    components.'''
    def __init__(self, begin_update, predict=None, predict_one=None,
            nI=None, nO=None, *args, **kwargs):
        self.begin_update = begin_update
        if predict is not None:
            self.predict = predict
        if predict_one is not None:
            self.predict_one = predict_one
        self.nI = nI
        self.nO = nO
        Model.__init__(self)
